#!/usr/bin/env python

import sys, os, subprocess
from commands import getoutput
from commands import getstatusoutput

### helper functions ###

def usage():
        print "Usage:"
        print 
        print sys.argv[0]+" < file | - >"
        print """
This is a python script. It takes a file or stdin as input 
assuming it is an Alice & Bob specification and
converts the specification into role scripts.
"""

class bcolors:
    BLACK = '\033[90m'
    RED = '\033[91m'
    BOLDRED = '\033[1;31m'
    YELLOW = '\033[93m'
    YELLOWRED = '\033[0;33;41m'
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = YELLOWRED
    FAIL = BOLDRED
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    ENDC = '\033[0m'

# set up to three attributes as follows:
# if all three are set it's in the order suggested by the variable names
# if fewer are set, BG is not being set
# if only one is set it can be an ATTR (e.g., bold, underlined) or a FG color.
def setcolor(ATTR,FG="",BG=""):
	    bcolor = "\033["
	    colors = { 'black': 0, 'red': 1, 'green': 2, 'yellow': 3, 'blue': 4, 'magenta': 5, 'cyan': 6, 'white': 7 }
	    attributes = { 'bold': "1", "dim": "2", 'underline': "3", 'blink': "4", 'reverse': "7", 'hidden': "8" }

	    if ATTR in attributes:
		    bcolor += attributes[ATTR]
		    if FG in colors:
			    bcolor += ";"+str(colors[FG]+30)
			    if BG in colors:
				    bcolor += ";"+str(colors[BG]+40)
		    bcolor += "m"
	    elif ATTR in colors:
		    bcolor += "0;"+str(colors[ATTR]+30)
		    if FG in colors:
			    bcolor += ";"+str(colors[FG]+40)
		    bcolor += "m"
	    else: 
		    bcolor += "0m"
	    return bcolor

def unsetcolor():
	    return bcolors.ENDC


# write an error on the stderr and quit the program 
def error(text):
	sys.stdout.flush()
	sys.stderr.write(bcolors.FAIL+"Error:"+bcolors.ENDC+" "+bcolors.BOLD+text+bcolors.ENDC+"\n")
	sys.stderr.flush()
	assert False
#        sys.exit(1)

def warn(text):
	sys.stdout.flush()
	sys.stderr.write(bcolors.WARNING+"Warning:"+bcolors.ENDC+" "+text+"\n")
#	sys.stderr.write(setcolor("bold","red")+"Warning:"+unsetcolor()+" "+text+"\n")
	sys.stderr.flush()
#        sys.exit(1)
    
def nowhitespace(m):
        r = m.replace(" ","")
        r = r.replace("\t","")
        r = r.replace("\r","")
        return r.replace("\n","")

# a useful string of characters
alphanum = "".join([ chr(i)+chr(i).lower() for i in range(ord("A"),ord("Z")+1)])+"0123456789"
    
# returns the maximal subsequence of s that consists only of characters in keep
def keep_char(s,keep):
	return "".join([ r for r in s if r in keep ])

# replaces multiple whitespace and tabs by single space
def singlewhitespace(m,whites=["\t"," "]):
        line = m
        for w in whites:
            line = " ".join(filter(lambda x: x != '',line.split(w)))
    
        line = line.strip(" ") # remove leading and trailing space
        return line

# return true if list a is a prefix of list b
# otherwise return false
def prefix(a,b):
        if len(a)>len(b):
            return False
        for i in range(len(a)):
            if a[i] != b[i]:
                return False
        return True

# copy a list of lists
def unshared_copy(inList):
        if isinstance(inList, list):
            return list( map(unshared_copy, inList) )
        return inList
    
    
# removes all symbols except for alphanumeric ones
def verbs(m):
        r = m
        tospaces = "\t\n\r[];:<>/.(),{}|+=-^#@!"
        for s in tospaces:
            r = r.replace(s," ")
        s = r.replace("  "," ")
        while s != r:
            r = s
            s = r.replace("  "," ")
        r = r.strip(" ")
        return r
    
# returns everything which looks like an alphanumberic function symbol, 
# separated by spaces
def functionsymbols(m):
        r = m+" "
        tospaces="\t\n\r[];:<>)}/.,|+=-^#@!"
        for s in tospaces:
            r = r.replace(s," ")
    
        # remove all names which are not function names, recognized by the 
        # fact that they continue with a space instead of "(" or "{"
        t = ""
        while t != r:
            t = r
            for s in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'\"":
                r = r.replace(s+" "," ")
    
        r = r.replace("("," ")
        r = r.replace("{"," ")
        s = r.replace("  "," ")
        while s != r:
            r = s
            s = r.replace("  "," ")
        r = r.strip(" ")
        return r
    
def getVarsFromList(l):
        r = []
        for m in l:
            r += getVariables(m)
        r = list(set(r))
        return r
    
# return all variable names in string m
def getVariables(m):
        l = []
        fsym = functionsymbols(m).split(" ")
        for s in verbs(m).split(" "):
            if (s not in fsym):
                l.append(s)
        return l

# removes duplicate entries from list of positions 
# and is order preserving
def unique(seq, idfun=None): 
   # order preserving
   if idfun is None:
       def idfun(x): return x
   seen = {}
   result = []
   for item in seq:
       marker = idfun(item)
       # in old Python versions:
       # if seen.has_key(marker)
       # but in new ones:
       if marker in seen: continue
       seen[marker] = 1
       result.append(item)
   return result

# sorts and ensures uniqueness in the positions in l
def sortedpos(l):
	return sorted(unique(l,tuple),cmp=position_compare)

# Compares two positions x and y (lists of positive integers)
# x < y  if the first position that x and y differ at is smaller in x than y
# or if x is a prefix of y 
def position_compare(x,y):
        for i in range(min(len(x),len(y))):
		if x[i] < y[i]: return -1
		elif x[i] > y[i]: return 1
	if len(x) < len(y): return -1
	elif len(x) > len(y): return 1
	else: return 0

#!/usr/bin/env python


class AnB: 

    debug = False

    def __init__(self):
#	    self.terms = Term()
            self.rolenames = [] # role names in order of appearance
            self.T = TermAlgebra() # Create a new Term Algebra T
	    self.prot = Protocol(self.T,"P")

#### The following is for debugging:
#            T.Sigma = { 'h': 1, 'enc': 2, 'dec':2, 'senc': 2, 'sdec':2, 'aenc': 2, 'adec': 2, 'list': -1, 'true': 0, 'pk': 1, 'fst': 1, 'snd': 1, 'pair':2  }
#            T.Equations = [ "dec(enc(%M,%K),%K)=%M", "adec(aenc(%M,pk(%K)),%K)=%M", "sdec(senc(%M,%K),%K)=%M", "fst(pair(%X,%Y))=%X", "snd(pair(%X,%Y))=%Y" ]

# add a new function symbol to the term algebra T
    def addFunction(self,f,arity):
        self.T.addFunction(f,arity)

    def process_line(self,line):
        if ":" not in line:
            if "knows" in line:
                self.process_line(line.replace("knows",": knows"))
            elif "fresh" in line:
                self.process_line(line.replace("fresh",": fresh"))
            elif "=" in line:
                lhs, rhs = line.split("=")
                lhs = Term(self.T,nowhitespace(lhs))
                rhs = Term(self.T,nowhitespace(rhs))
                vars = lhs.vars()+rhs.vars()
                vars = list(set(vars))
                if self.debug:
                    print "Left-hand side: "+lhs.to_str()
                    print "Right-hand side: "+rhs.to_str()
                    print "Variables: ",vars
                u = {}
                for v in vars:
                    u[v] = "%"+v
                lhs = lhs.apply(u)
                rhs = rhs.apply(u)
                if self.debug:
                    print "Resulting equation: "
                    print lhs.to_str()+"="+rhs.to_str()
                self.T.addEquation(lhs.to_str()+"="+rhs.to_str())
                # read equations line by line, substitute %variable for every variable
            else:
                error("Invalid input: >>"+line+"<<\n")

        elif nowhitespace(line).lower().startswith("functions:"):
            funcs = line.split(":",1)
            for fs in funcs[1].split(","):
                if self.debug: print "Splitting: ",fs
                if nowhitespace(fs) != "":
                    f,arity = fs.split("/",1)
                    f = nowhitespace(f)
                    arity = int(arity)
                    if self.debug: print "Adding: "+f+" with arity "+str(arity)
                    self.addFunction(f,arity)
        elif nowhitespace(line).lower().startswith("equations:"):
            line = line.split(":",1)[1]
            self.process_line(line)
        else:
            AB, m = line.split(":", 1)
            if "->" in AB:
                A,B = AB.split("->", 1)
                A = nowhitespace(A)
                B = nowhitespace(B)
                if A not in self.rolenames:
                    self.rolenames.append(A)
                if B not in self.rolenames:
                    self.rolenames.append(B)
                for rname in self.rolenames:
                    self.prot.add_role(rname,self.rolenames)
    
                self.prot.add_event(B,"receive",{ 'msg': m.strip(" "), 'from': A})
                fsym = functionsymbols(m).split(" ")
                for s in fsym:
                    if s != "" and s not in self.T.Sigma:
                        warn("Unknown function symbol >"+s+"<")
                        self.addFunction(s,-1)
    
                self.prot.add_event(A,"send",{ 'msg': m.strip(" "), 'to': B})
    
            else:
                AB = nowhitespace(AB)
                m = nowhitespace(m)
                if m.startswith("fresh"):
                    stuff = m.split("fresh",1)[1]
                    if stuff[0] == "(" and stuff[-1] == ")":
                        stuff = stuff[1:-1]
                    fsym = functionsymbols(stuff).split(" ")
                    for s in fsym:
                        if s != "":
                            warn("Keyword fresh applied to function symbol >"+s+"<")
                    for i in Term(self.T,"_list("+stuff+")").arguments():
                        if i.is_variable():
                            self.prot.add_event(AB,"fresh",{'msg': nowhitespace(i.to_str())})
                        else:
                            warn("Error fresh keyword applied to a term which is not a variable: "+i.to_str())

                elif m.startswith("knows"):
                    stuff = m.split("knows",1)[1]
                    if stuff[0] == "(" and stuff[-1] == ")":
                        stuff = stuff[1:-1]
                    fsym = functionsymbols(stuff).split(" ")
                    for s in fsym:
                        if s != "" and s not in self.T.Sigma:
                            warn("Unknown function symbol >"+s+"<")
                            self.addFunction(s,-1)

                    knowstuff = []
                    for i in Term(self.T,"_list("+stuff+")").arguments():
                        knowstuff.append(nowhitespace(i.to_str()))
                    self.prot.add_role(AB,knowstuff)
                else:
                    sys.stderr.write("Invalid input: >>"+line+"<<\n")
                    sys.exit(1)
    
    
    def read_from_file(self,arglist):
        filename = ""
        for arg in arglist:
            if arg == "-h" or arg == "--help":
                usage()
                sys.exit(1)
            else: 
                try:
                    f = open(arg)
                            #        print "# Reading from "+arglist[1]
                    filename = arg # .split(".",1)[0]
                except:
                    sys.stderr.write("Unrecognized command line argument.\n")
                    sys.stderr.write("Couldn't open >>"+arg+"<< for reading.\n")
    
        if filename == "":
            filename = "stdin"
            f = sys.stdin

        if filename != "stdin":
            sanitized_filename = nowhitespace(filename).replace("-","_")
            sanitized_filename = sanitized_filename.split("/")[-1]
            sanitized_filename = sanitized_filename.split("\\")[-1]
            if "." in sanitized_filename:
                sanitized_filename = sanitized_filename.split(".")[-2]

            self.prot.protocol_name = keep_char(sanitized_filename,alphanum+"_")
        else:
            self.prot.protocol_name = "noname"

        try:
            for line in f:
                line = line.rstrip('\n')  # this is the perl chomp()
                # The following are considered comments
                line = line.split("//",1)[0]
                line = line.split("#",1)[0]
                line = line.split("%",1)[0]
    
                # remove white space
                line = "".join(line.split(" "))
                line = "".join(line.split("\t"))
                
                # 
                if line != "":
                    self.process_line(line)
    
        finally:
            if type(f) != type([]):
                f.close()
    
    def read_prot(self,arguments):
        self.read_from_file(arguments)

    def output(self):
        self.prot.output()

    def output_AnB(self):
        events = {}
        roles = self.prot.roles
        for role in roles:
            events[role] = roles[role].firstEvent

        buffer = [] # collects all events which were generated as a consequence of a particular A&B specification line

        nochange = False
        while len(events.keys()) > 0:
            if nochange:
                warn("Looping without progress. Exiting.")
                break
            nochange = True
            for role in events.keys():
                if events[role] == None:
                    del events[role]
                    nochange = False
                elif events[role].action == "knows":
                    print role+": knows ",events[role].arguments['msg']
                    events[role] = events[role].next
                    nochange = False
                elif events[role].action == "fresh":
                    print role+": fresh ",events[role].arguments['msg']
                    events[role] = events[role].next
                    nochange = False
                elif events[role].action == "send":
                    buffer.append([role,events[role]])
                    events[role] = events[role].next
                    nochange = False
                elif events[role].action == "receive":
                    msg = events[role].arguments['msg']
                    for i in range(len(buffer)):
                        r = buffer[i][0]
                        e = buffer[i][1]
                        if e.arguments['msg'] == msg:
                            print r+" -> "+role+": "+msg
                            del buffer[i]
                            events[role] = events[role].next
                            nochange = False
                            break
                else:
                    warn("Unknown action: "+events[role].action)
                    nochange = False
                    events[role] = events[role].next


#### end of class AnB ####

### main:

if __name__ == "__nomainhere__":

    arguments = sys.argv[1:]

    ABC = AnB()

## Parse the Alice & Bob specification into a protocol type
    ABC.read_prot(arguments)

## Output the protocol
    ABC.output()
#!/usr/bin/env python



## Knowledge stores terms (of type Term) and uses TermAlgebra to derive new terms from known terms
class Knowledge:
    debug = False

    def __init__(self,T):
        self.KD = {}
	self.T = T     # TermAlgebra

    def klistup(self,l):
        uni = [ {} ]
        for t in l:
            uniprime = []
            for u in uni:
                uniprime += self.kup(t.apply(u))
            uni = uniprime
        if uni == []:
            return False
        return True
        
           
    def kup(self,t):
        if t.to_str() in self.KD:  # if t has been constructed before, we're done
            return [ {} ]
        elif t.is_ground() and t.is_variable(): # if t is an atomic term (and not in KU, not constructable) then we're doomed. FIX: rename 'var' to 'atomic'
            return [ ]
        elif t.is_variable(): # t is a variable, can be anything
            return [ {} ]
        else:
            # t cannot be obtained by deconstuction, perhaps subpieces of it can: try to construct it from pieces:
            l = []  
            for i in t.arguments(): # t.term['arg'] (but typed as Term)
                r = self.kup(i)
                if self.debug: print "Unifier(s) received: ",r
                lres = []
                for u1 in l:
                    for u2 in r:
                        u3 = combinesub(self.T,u1,u2)
                        if u3 != False:
                            lres.append(u3)
            unifiers = []
            for u in lres:
                for c in self.KD:
                    u = mgusub(self.T,c,t.apply(u).to_str())
                    if unifies(u):
                        unifiers.append(u)
            return unifiers
    
    # To find a fixed point in KD
    # 1. Given a term t, drules = [{kdl,[ku1,ku2,...],kdr},...]
    # 2. find applicable rules by trying to unify t with kdl
    # 3a. eliminate rules that yield a kdr which is already in KD
    # 3b. reduce applicable rules to those that have unifiers with KU
    # 4. obtain kdr, add it.
    
    # KU should be closed to contain smallest elements necessary    
    def closeKD(self):
    #    KDmax = 0
    #    for t in KD.keys():
    #        s = msgsize(t) 
    #        if s > KDmax:
    #            KDmax = s
        if self.debug:
		print "(closeKD) T.printDrules:"
		self.T.printDrules()
        again = True
        while again:
            again = False
	    if self.debug:
		    print "Knowledge: "
		    print self.KD.keys()
            # try to deconstruct term m
            for m in self.KD.keys():
                t = self.KD[m]
		if not t.is_variable():  # we cannot deconstruct vars
                  # for every deconstruction rule
                  for r in self.T.drules:
   		    if self.debug: 
			print "Calling mgusub with m = "+m+" r['ld']="+r['ld'].to_str()
		    # check to see whether the rule r applies to term m
                    u = mgusub(self.T,m,r['ld'].to_str())
                    if unifies(u):
			# r applies to m, now apply the same unifier
			# to the KU terms (listed in r['lc']) to apply the rule
                        l = []
                        for rt in r['lc']:
                            l.append(rt.apply(u))
			# see whether we already know
			# the result of applying the rule to m
                        rhs = r['r'].apply(u)
                        if rhs.to_str() not in self.KD:
                            # we don't know the result, check that we can
                            # actually construct the necessary KU terms
                            if self.klistup(l):  
 			        # yes, we cann, add it to our knowledge
                                self.add(rhs)
                                # we have learned something new, so do 
                                # all of this again 
                                again = True
    
    # checks whether a term is derivable from the knowledge.
    # starts by closing the knowledge, then checks constructability
    def is_derivable(self,t):
	    self.closeKD()
            return self.is_constructible(t)
    
    # This will be more useful when annotations need to be added, too
    def add(self,m):
        if type(m) is str:
		if m not in self.KD:
			self.add(Term(self.T,m))
		return
        s = m.to_str()
        if s not in self.KD:
            self.KD[s] = m
    
    # return a list of terms in KD (known terms)
    # the list is a copy of these terms (perhaps [:] is not necessary?)
    def list(self):
	    return self.KD.keys()[:]

    # tells whether a term can be constructed from subterms, purely
    # by applying functions to subterms, no deconstructions!
    # see also: is_derivable above
    def is_constructible(self,t):
        if t.to_str() in self.KD:
            return True  
        elif t.is_variable():
            return False
        else:
        # try to construct it from pieces:
            for i in t.arguments():
                if not self.is_constructible(i):
                    return False
            return True
    
    def minimalKnowledgeList(self):
	    minimal = []
	    for ki in self.KD:
#		    print "Minimizing ki = "+ki
		    t = Term(self.T,ki)
		    if t.is_atomic():
#			    print ki+" is atomic, adding to minimal knowledge"
			    minimal.append(ki)
		    else:
		        for kj in t.arguments_as_strlist():
			    if not self.is_constructible(Term(self.T,kj)):
#				    print kj+" is not a known subterm of "+ki+" adding ki to minimal knowledge"
				    minimal.append(ki)
				    break
	    return minimal

    def minimalKnowledge(self):
	    minimal = []
	    for ki in self.KD:
#		    print "Minimizing ki = "+ki
		    t = Term(self.T,ki)
		    if t.is_atomic():
#			    print ki+" is atomic, adding to minimal knowledge"
			    minimal.append(t)
		    else:
		        for kj in t.arguments_as_strlist():
			    if not self.is_constructible(Term(self.T,kj)):
#				    print kj+" is not a known subterm of "+ki+" adding ki to minimal knowledge"
				    minimal.append(t)
				    break
	    return minimal

    def printKnowledge(self):
        for i in self.KD:
            print "  "+i,
        print
    

####### MAIN ##########
if __name__ == "__nomainhere__":

   arguments = sys.argv[1:]
#   if len(arguments) >= 1 and arguments[0][0]=="-":
#	    print "first argument: ",arguments[0]
#	    arguments = arguments[1:]
#   else:
#        usage()
#        print
#        sys.stderr.write("Error: No output format specified.\n")
#        sys.exit(0)

   ABC = AnB()
#   ABC = strand()

## Parse the Alice & Bob specification into a protocol type
   ABC.read_prot(arguments)

## Output the protocol
   ABC.output()

## Go through the protocol specification and print knowledge
   for r in ABC.prot.roles:
       print "Role "+ABC.prot.roles[r].role_name
       K = ABC.prot.roles[r].IK
       ev = ABC.prot.roles[r].firstEvent
       step = 0
       while ev != None:
	   step += 1
	   action = ev.action
	   arg = ev.arguments
	   ev = ev.next
	   if action == "receive":
                m = Term(K.T,arg['msg'])
                print str(step)+". Receiving: "+m.to_str()
                K.add(m)
                K.closeKD()
		if K.debug:
			K.printKnowledge()
	   elif action == "fresh":
                m = Term(ABC.T,arg['msg'])
                print str(step)+". Fresh: "+m.to_str()
                K.add(m)
                K.closeKD()
		if K.debug:
			K.printKnowledge()		   
	   elif action == "send": 
                m = Term(K.T,arg['msg'])
                print str(step)+". Sending: "+m.to_str()
                if K.debug:
                    print "Initial knowledge:"
                    for i in KD:
                        print i,
                    print
                    K.closeKD()  # not necessary since done when receiving messages
                
                    print "Closed knowledge:"
                    for i in K.KD:
                        print i,
                    print
                if K.is_constructible(m):
			if K.debug: print m.to_str()+" is derivable"
                else:
                    warn(m.to_str()+" is not derivable")
	   else:
                if 'msg' in arg:
			m = Term(K.T,arg['msg'])
			m = m.to_str()
		else:
			m = str(arg)
                print str(step)+". unknown action '"+action+"' with argument "+m
#                K.add(m)
                if K.debug:
                    print "Closed knowledge:"
                    for i in K.KD:
                        print i,
                    print
    

maude_path = ""

debug = False

# recall: spaces are significant in maude!!! make spaces.
def maude_fmod(T,with_equations=False):
    Sigma = T.Sigma
    maudefunc = ""
    maudeeq = ""
    for f in Sigma:
        maudefunc += "op "+f+" : "+"A "*Sigma[f]+" -> A .\n"
    maudestr = """fmod MSG is
protecting QID .
sort A .
subsort Qid < A .
"""
### A test module:
# fmod MSG is
#  sort A .
#  op fst : A -> A .
#  op snd : A -> A .
#  op < _, _ > : A A -> A .
#
#  vars a b : A .
#  eq fst(<a,b>) = a . 
#  eq snd(<a,b>) = b . 
# endfm
# """
    if with_equations: # We don't need equations for syntactic unification 
        vars = []
        for eq in T.Equations:
            eq = eq.replace("%","")
            lhs,rhs = eq.split("=")
            lhs = Term(T,lhs)
            rhs = Term(T,rhs)
            vars = list(set(vars+lhs.vars()+rhs.vars()))
            maudeeq += "eq "+msg2maudestr(lhs.term)+" = "+msg2maudestr(rhs.term)+" .\n"

# We don't want predefined variables in the module. 
#        if maudeeq != "":
#            maudeeq = "vars "+" ".join(vars)+" : A .\n\n"+maudeeq

    maudestr += maudefunc+maudeeq+"\nendfm\n"
    if debug: print "(maude_fmod) Function module: "+maudestr

    return maudestr


# reduce a term - could perhaps be useful to rewrite {a}k to enc(a,k)
def maude_simplify_command(T,term):
    maudestr = str2maudestr(T,term)
    maudestr = "reduce in MSG : "+maudestr+" .\n"
    maudestr += "quit()\n"
    return maudestr

def maude_simplify_result(output):
    # need to handle the output of maude here
    [ garbage, result ] = output.split("result A:")
    [ result, garbage ] = result.split("\n")
    result = cleanstr(result)
    print "(maude_simplify) Result: "+result
    return result


def maude_do_via_file(maudestr):
    try:
        f = open("MSG.temp.maude","w")
        f.write(maudestr)
    except:
        error("Preparation for maude interaction failed.")
    finally:
        f.close()

    # Can call maude with "maude file", where file contains all 
    # the queries and then ends with a quit()
    # or can call it with "maude < file", where file contains all 
    # the queries and no quit()
    # The latter will print "Maude> " prompts, the former won't. 
    # The latter might be preferable for intereactive sessions though?
    status, output = getstatusoutput(maude_path+"maude -batch -no-banner -no-ansi-color MSG.temp.maude")
    print "(maude_do) Result: "+output
    return output

def maude_do_via_pipe(maudecommand):
    proc = subprocess.Popen(["maude","-no-wrap","-batch","-no-banner","-no-ansi-color"],stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.PIPE)
    proc.stdin.write(maudecommand)
    proc.stdin.close()
    r = proc.stdout.read()
    s = proc.stderr.read()
    if "Warning" in s:
        error("Error in the following command to maude:\n"+maudecommand)
    elif len(s) > 0:
        warn("maude - stderr: "+s)
    if debug: 
        print "(maude_do_via_pipe) Maude Subprocess returned on stdout: ",r
        print "(maude_do_via_pipe) Maude Subprocess returned on stderr: ",s
    proc.wait() 
    return r

def maude_do(maudecommand):
    return maude_do_via_pipe(maudecommand)

def msg2maudestr(msg):
    s = ""
    if msg['type'] == 'var': # and not ground(msg):
        if msg['name'][0] == "%":
            s += msg['name']+":A"
        else:
            s += "'"+msg['name']
    else:
        s = msg['name']
    if msg['type'] == 'func':
        s += "("
        for i in range(msg['arity']):
            if i > 0:
                s += ","
            s += msg2maudestr(msg['arg'][i])
        s += ")"
    return s

def maudestr2msg(s):
    r = s.replace(":A","")
    r = r.replace("'","")
    return r

def str2maudestr(T,s):
    m = Term(T,s)
    return msg2maudestr(m.term)

def maude_unify_command(T,s1,s2):
    m1 = str2maudestr(T,s1)
    m2 = str2maudestr(T,s2)
    # If we have several things to unify simultaneously we can do better than the following. We should then give all those things to maude connected with /\
    command = "unify in MSG : "+m1+" =? "+m2+" ."
    command += "\nquit()\n"
    return command

def maude_unify_result(r):
    # the following does not work if there are no solutions
#    [ garbage, result ] = r.split("Solution 1")
# I believe that decision time is the last uninteresting line
    [ garbage, result ] = r.split("Decision time:")
# now remove all up to the first newline 
    [ garbage, result ] = result.split("\n",1)

    [ result, garbage ] = result.split("Bye.")

    # finally remove Solution 1 if it occurs:
    if "Solution 1" in result:
        [ garbage, result ] = result.split("Solution 1")

    r = result
    r = r.lstrip("\n")
    if debug: print "(maude_unify_result) Cleaned result: "+r+"."
    return r

# unifier tool:
def mgusub(T,s1,s2):
    rho = {}
    # maude_fmod() will need to take Sigma and "ground" terms as argument
    maudecommand = maude_fmod(T)+maude_unify_command(T,s1,s2)
    
    r = maude_do(maudecommand)
    r = maude_unify_result(r)

    if "No unifier" in r:
        if debug: print "(mgusub) "+s1+" and "+s2+" are not unifiable"
        return False
    elif "Solution 2" in r:
        if debug: print "(mgusub) "+s1+" and "+s2+" have more than one solution!"
        error("Unexpected result in mgusub.")
        
    if debug: print "(mgusub) "+s1+" and "+s2+" unify via "+r

    substlist = r.split("\n")

    pre_rho = {}  # contains maude's #1:A, #2:A, .... variables
    # clean the strings: replace #1:A, ... by the first variable name 
    # that is assigned to them, later remove :A
    for i in range(len(substlist)):
        s = substlist[i]
        if " --> " in s:
            lhs,rhs = s.split(" --> ")
            lhs = nowhitespace(lhs)
            rhs = nowhitespace(rhs)
            if rhs[0] == "#":
                if rhs not in pre_rho:
                    pre_rho[rhs] = lhs
                else:
                    rho[maudestr2msg(lhs)] = maudestr2msg(pre_rho[rhs])
                # remove this entry from substlist
                substlist[i] = ""

    for s in substlist:
        if " --> " in s:
            lhs,rhs = s.split(" --> ")
            lhs = nowhitespace(lhs)
            rhs = nowhitespace(rhs)
            if rhs[0] == "#":
                error("(mgu-sub) This should not be happening, # have been remvoed.")
            else:
                for pr in pre_rho:
                    if pr in rhs:
                        rhs = rhs.replace(pr,pre_rho[pr])
                if "#" in rhs:
                    warn("Debug info for upcoming error:")
                    print "pre-rho: ",pre_rho
                    print "rho: ",rho
                    error("Could not remove # from the following substitutions:\n"+r)
        # remove :A 
                lhs = maudestr2msg(lhs)
                rhs = maudestr2msg(rhs)
                if lhs in rho:
                    print "(mgusub) That should not happen: "+lhs+" => "+rhs+", where lhs already in rho, evaluating to "+rho[lhs]
                rho[lhs] = rhs
        else:
            if s != "":
                warn("(mgusub) Unexpected line: "+s)
    return rho

def assign(term,assignment):
    debug = True
    subterm = deconstruct(term)
    for j in range(len(subterm)):
        if subterm[j] in assignment:
            if debug: print "(assign) subterm "+subterm[j]+" is now assigned to ",
            subterm[j] = assignment[subterm[j]]
            if debug: print subterm[j]
        else:
            if debug: print "(assign) subterm "+subterm[j]+" is not a variable."
            
    newterm = "".join(subterm)
    if debug: print "(assign) after substitutions we have "+newterm
    return newterm

def combinesub(T,rho,sigma):
    debug = True
    if (rho == False) or (sigma == False):
        return False
    # Make a copy of sigma
    u = {}
    for s in sigma:
        u[s] = sigma[s]

    for s in rho:
        if s in u:
            if debug: print "(updatesub) Computing mgu of two subs"
            localsubs = mgusub(T,u[s],rho[s])
            if localsubs == False:
                if debug: print "(updatesub) Not unifiable substitutions: "+s+" => "+sigma[s]+" vs "+rho[s]
                return False
            else:
                u[s] = assign(u[s],localsubs)
        else:
            u[s] = rho[s]
    return u


def unifies(u):
    if u == False:
        return False
    else:
        return True


####### MAIN ##########
if __name__ == "__nomainhere__":

    Term._Sigma_ = TermAlgebra._Sigma_
    print mgusub("enc(a,b)","enc(b,a)")
    print mgusub("enc(a,b)","x")
    print mgusub("enc(a,b)","h(b)")
#!/usr/bin/env python


class Term:
    # this class converts strings to the protocol message type:
    # a protocol message is a dictionary: ('type','name','arity','arg'),
    # where type is one of 'func', 'var'
    #       name is the name of the function, or the variable
    #       arity is used for functions: 0 constant, positive, negative: usef for lists
    #       list contains further dictionaries, i.e. every list element is a msg

    debug = False

    # T is an instance of the TermAlgebra class
    def __init__(self,T,msgstr = ""):
	    if self.debug: print "Setting up msgstr: ",msgstr
            self.T = T
	    if type(msgstr) is str:
		    self.set(msgstr)
	    elif type(msgstr) is dict:
		    self.term = msgstr
	    elif type(msgstr) is Term:
		    print "That's unexpected."
		    error("Type mismatch type is "+str(type(msgstr)))
	    else:
		    print "Term __init__ received: ",msgstr
		    error("Type mismatch type is "+str(type(msgstr)))


    def set(self,s):
        self.term = self._str2msg(s)

    def information(self):
        print "TermAlgebra: ",self.T
        print "Prettyprint: "+self.to_str()
        print "Variables: "+str(self.vars())
	print "Ground term: ",self.is_ground()
#        print "Position [1,1,2]: "+self.position([1,1,2]).to_str()


    def top_name(self):
	    return self.term['name']

    def is_atomic(self):
        return (self.is_variable() or len(self.term['arg']) == 0)

    def is_variable(self):
	    if self.term['type'] == 'var':
		    return True
	    return False

    def is_ground(self):
	    return self._ground(self.term)

    def _ground(self,t):
        if t['type'] == 'func':
            for i in t['arg']:
                if not self._ground(i):
                    if self.debug: print i['name']+" is not ground"
                    return False
            if self.debug: print t['name']+" is ground"
            return True
        else:
            if t['name'][0] in "%": # "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
                if self.debug: print t['name']+" is not ground"
                return False
            if self.debug: print t['name']+" is ground"
            return True

    # return a list of Term-typed elements. These elements are the arguments of the top-level function. 
    # They are not copies! 
    # If the term given does not have a top-level function, the empty list is returned.
    def arguments(self):
	    l = []
            for i in self.term['arg']:
		    l.append(Term(self.T,i))  # This is not a copy
	    return l

    def arguments_as_strlist(self):
	    l = []
            for i in self.term['arg']:
		    l.append(Term(self.T,i).to_str())  # This is not a copy
	    return l

    # return a list of all positions in term that are the children of position l
    # A position is a list of positive integers
    def argument_positions(self,l = []):
        return [ l+[i+1] for i in range(self._msgposition(self.term,l)['arity']) ] 

    # apply a substitution to term. Returns a copy of term with the substitution applied
    # The substituion u is a dictionary mapping strings to strings. The keys of u are names 
    # of variables. The values are the terms (as strings) the variables are assigned to.
    def apply(self,u):
	    m = Term(self.T,self._apply(u,self.term))
	    #    print "Information on subst application: "
	    #    m.information()
	    return m

    def _apply(self,u,t):
        m = {}
        if t['type'] == 'func':
            m['name'] = t['name']
            m['type'] = t['type']
            m['arity'] = t['arity']
            m['arg'] = []
            for i in t['arg']:
                m['arg'].append(self._apply(u,i))
        else:
            if t['name'] in u:
                m = self._str2msg(u[t['name']])
            else:
                m['name'] = t['name']
                m['type'] = t['type']
                m['arity'] = t['arity']
                m['arg'] = unshared_copy(t['arg'])  
        return m
    
    # replaces position pos in self.term by newterm
    # note: newterm is not copied! 
    def replace_in_pos(self,pos,newterm):
        if pos == []:
            self.term = newterm.term
        else:
            m = self.term
            for i in pos[:-1]:
                if m['arity'] >= i:
                    m = m['arg'][i-1]
                else:
                    error("There is no subterm at position "+str(l))
            i = pos[-1]
            if m['arity'] >= i:
                m['arg'][i-1] = newterm.term

    def msg_variable(self,name):
        if name not in self.T.Var:
            self.T.Var[name] = 1
    
        return {'type': 'var', 'name': name,'arity': 0,'arg': []}
    
    def msg_func(self,name,s):
        Sigma = dict(self.T.Sigma)
        Sigma['_list'] = -1 # add a function called _list. To be used internally only
        if name in Sigma:
            l = self.msg_list(s)
            if l['arity'] == Sigma[name] or Sigma[name] == -1:
                return {'type': 'func','name': name, 'arity': l['arity'],'arg': l['arg']}
            else:
                error("Number of arguments ("+str(l['arity'])+") does not match function definition. "+name+" takes "+str(Sigma[name])+" arguments.")
        else:
            error("Unknown function: "+name)
    
    def msg_list(self,s):
        l = self.str2args(s)
        arg = []
        larg = "("
        rarg = ")"
        for i in l:
            if larg in i:
                name,remainder = i.split(larg,1)
                # remainder should end with rarg
                if remainder[-1] != rarg:
                    error("Syntax error. Expected "+rarg+" as last character in "+remainder)
                arg.append(self.msg_func(name,remainder[:-1]))
            else:
                arg.append(self.msg_variable(i))
    
        return {'type': 'func', 'name': '_list', 'arity': len(arg), 'arg': arg}
    
    # returns list of arguments of the top-level function in string s
    # arguments in list are strings
    def str2args(self,s):
        if len(s) == 0:
            error("class Term, function str2args: empty list")
        l = []
        c = ""
        i = 0
        larg = "("
        rarg = ")"
        sep = ","
        numparen = 0
        while i < len(s):
            if numparen == 0 and s[i] == sep:
                l.append(c)
                c = ""
            else:
                c += s[i]
            
            if s[i] == larg:
                numparen += 1
                i += 1
            elif s[i] == rarg:
                numparen -= 1
                i += 1
    #        elif s[i] == esc:
    #            i += 2
            else:
                i += 1
        l.append(c)
        if numparen != 0:
            error("Unbalanced parentheses in string "+s)
    
    #    print "Returning: ",l
        return l
    
    # convert a string to a term
    def _str2msg(self,s):
        m = self.msg_list(s)
        if m['arity'] == 1:
            m = self._msgposition(m,[1])
        else: 
            warn("Class Term, _str2msg: Received list of terms instead of a single term")
        return m

    # return the list of variables
    def vars(self):
	    return self.msgvars(self.term)

    # return the list of variables in msg
    def msgvars(self,msg):
        l = []
        if msg['type'] == 'func':
            for i in range(msg['arity']):
                l += self.msgvars(msg['arg'][i])
        elif msg['type'] == 'var':
            l.append(msg['name'])
        return l

    # return the list of all names
    def names(self):
        return self.msgnames(self.term)
    
    # return the list of all names in msg (term)
    def msgnames(self,msg):
        l = []
        if msg['type'] == 'func':
            for i in range(msg['arity']):
                l += self.msgvars(msg['arg'][i])
        l.append(msg['name'])
        return l

    # return a list of all positions in term below position l
    def positions(self,l = []):
        return self.msgpositions([self._msgposition(self.term,l)],[l])

    def msgpositions(self,msgs,l):
        newterms = []
        newpositions = []
        for i in range(len(msgs)):
            for j in range(len(msgs[i]['arg'])):
                newterms.append(msgs[i]['arg'][j])
                newpositions.append(l[i]+[j+1])
        if newpositions != []:
            newpositions = self.msgpositions(newterms,newpositions)
        return l+newpositions

    # return the subterm at position l
    # positions are positive integers
    def position(self,l):
	    return self.msgposition(self.term,l)

    def msgposition(self,msg,l):
        return Term(self.T,self._msgposition(msg,l)) ### this is not a copy

    # return the subterm of term msg at position l
    # positions are positive integers
    def _msgposition(self,msg,l):
        m = msg
        for i in l:
            if m['arity'] >= i:
                m = m['arg'][i-1]
            else:
                error("There is no subterm at position "+str(l))
        return m
    
    def to_str(self):
	    return self.msg2str(self.term)

    def msg2str(self,msg):
        s = msg['name']
        if msg['type'] == 'func':
            s += "("
            for i in range(msg['arity']):
                if i > 0:
                    s += ","
                s += self.msg2str(msg['arg'][i])
            s += ")"
        return s    

    # find all positions of term s in this term
    def find(self,s):
        return self.findstr(self.msg2str(s.term))
    
    # called by find(s).
    # Same as find(), but s here is a string
    def findstr(self,s):
	    return self._msgfindstr(s,self.term,[])

    # called by findstr(s,m), should not be called directly.
    # find all positions of s in msg, where s is a string and msg
    # is a term
    def _msgfindstr(self,s,msg,prefix):
        positions = []
        m = self.msg2str(msg)
        if s == m:
            return [prefix+[]]
        elif msg['type'] == 'var':
            return []
        else:
            for i in range(msg['arity']):
                positions += self._msgfindstr(s,msg['arg'][i],prefix+[i+1])
            return positions
    
    # siblings of the sub-term of this term at position p
    def siblings(self,p):
        siblings = []
        if p == []:
            return []
        parent = p[:-1]
        m = self.position(parent)
        sib = range(1,m.term['arity']+1)
        for i in sib:
            if i != p[-1]:
                siblings.append(parent+[i])
        return siblings
    
    # the size of this term
    def size(self):
	    return self.msgsize(self.term)

    def msgsize(self,msg):
        s = 1
        if msg['type'] == 'func':
            for i in msg['arg']:
                s += self.msgsize(i)
        return s
    
    

#### end of class terms ####
#!/usr/bin/env python


class TermAlgebra:
    debug = False
    debugmore = False
    _Sigma_ = { 'h': 1, 'enc': 2, 'dec':2, 'senc': 2, 'sdec':2, 'list': -1, 'true': 0, 'pk': 1, 'f': 3, 'fst': 1, 'snd': 1, 'pdec': 3, 'penc':3, 'pair':2  }
    _Var_ = {}
    _Equations_ = [ "dec(enc(%M,%K),%K)=%M", "fst(pair(%X,%Y))=%X", "snd(pair(%X,%Y))=%Y", "pdec(penc(%M,%K,%R),%K,%S)=%M" ]
    
    def __init__(self):
        self.Sigma = {} # self._Sigma_
        self.Var = {} # self._Var_
        self.Equations = [] # self._Equations_
        self.drules = []

        if self.debug: print "Sigma: ",self.Sigma

	self.setup_rules()


    def addFunction(self,f,arity):
        if f not in self.Sigma:
            self.Sigma[f] = arity

    def addEquation(self,eqn):
        self.Equations.append(eqn)
	self.setup_rules()

    def setup_rules(self):
    #### The following are obtained by following steps 1 through 4 in AnBfestschrift
        for arg in self.Equations:
            arg = nowhitespace(arg)
            ls,rs = arg.split("=")
            l = Term(self,ls)
            r = Term(self,rs)
    
            if self.debug: print "Eq: "+l.to_str()+" = "+r.to_str()
            if self.debugmore: print "size(l) = ",l.size(),", size(r) = ",r.size()
            if self.debug:
                print "All positions in l:"
                print l.positions()
            positions = l.find(r)
            if self.debugmore: print "set of positions: "+str(positions)
    
            for p in positions:
                if self.debugmore: print "Position: p = "+str(p)
                Dp = []
                for i in range(1,len(p)):
                    Dp.append(p[:i])
                if self.debugmore: print "Dp: ",Dp
    
                for pprime in Dp:
                    if self.debugmore: print "Position p' = "+str(pprime)
                    parent = pprime
                    Cp = []
                    while parent != []:
                        Cp += l.siblings(parent)
                        parent = parent[:-1]
                    if self.debugmore: print "Cp: ",Cp
                    
                    r = {}
                    r['lhseq'] = l
#                    r['rhseq'] = r  # not necessary since it is l.position(r['rp'])
                    r['ld'] = l.position(pprime)
                    r['r'] = l.position(p)
                    r['lc'] = []
                    r['rp'] = p # position of r in l
                    r['rdp'] = p[len(pprime):] # relative position of r in ld
                    if self.debugmore: print "Deconstruction rule:"
                    if self.debug: print "[ KD("+l.position(pprime).to_str()+")",
                    for q in Cp:
                        r['lc'].append(l.position(q))
                        if self.debug: print ", KU("+l.position(q).to_str()+")",
                    if self.debug:
                        print "]-->[",
                        print "KD("+l.position(p).to_str()+")",
                        print "]"
                        print
                        print "Replacement test: "
                        print "replacing "+l.position(r['rp']).to_str()
                        n = Term(self,l.to_str())
                        n.replace_in_pos(r['rp'],Term(self,"hithere"))
                        print n.to_str()
                    self.drules.append(r)
    #    print self.drules
 
# Var is still not being used for anything   
#        if self.debug:
#            print
#            print
#            print "Variables: ",self.Var

    def rule2str(self,r):
        s = "[ KD("+r['ld'].to_str()+")"
        for q in r['lc']:
            s += ", KU("+q.to_str()+")"
        s += " ]-->[ KD("+r['r'].to_str()+") ]"
        return s
    
    def printDrule(self,r):
        print self.rule2str(r)
    
    def printDrules(self):
        print "All Drules:"
        for r in self.drules:
            self.printDrule(r)


    
### main:

if __name__ == "__nomainhere__":

    arguments = sys.argv[1:]

    ABC = AnB()

## Parse the Alice & Bob specification into a protocol type
## and report on how the term algebra is used
    TermAlgebra.debug = True
    TermAlgebra.debugmore = True
    ABC.read_prot(arguments)


#!/usr/bin/env python


class Event:
    debug = False
    def __init__(self,action,arguments):
        self.action = action
        self.arguments = arguments
        self.next = None
#        self.prev = None

    def append(self,action,arguments):
        ev = Event(action,arguments)
        self.next = ev
#        ev.prev = self

    def insert(self,action,arguments):
        if self.debug: print "(insert) inserting "+action
        ev = Event(action,arguments)
        ev.next = self.next
#        ev.prev = self
        self.next = ev
#        if ev.next != None:
#            ev.next.prev = ev
        if self.debug: print "(insert) done"

    def insertBefore(self,action,arguments):
        error("This is deprecated and a stupid idea")

    def _insertBefore(self,action,arguments):
#        ev = Event(action,arguments)
#        ev.next = self
#        ev.prev = self.prev
#        self.prev = ev
#        if ev.prev != None:
#            ev.prev.next = ev
        if self.debug: print "(insertBefore) inserting "+action+" before "+self.action
        self.insert(action,arguments)
        tmparg = self.arguments
        tmpaction = self.action
        self.arguments = self.next.arguments
        self.action = self.next.action
        self.next.arguments = tmparg
        self.next.action = tmpaction
        if self.debug: print "(insertBefore) done"

    def add_arg(self,key,arg):
        if key not in self.arguments:
            self.arguments[key]=arg

    def update_arg(self,key,arg):
        self.arguments[key]=arg
            

class RoleScript:
    def __init__(self,T,role_name):
        self.role_name = role_name
        self.T = T # TermAlgebra
        self.IK = Knowledge(T) # initial knowledge
        self.firstEvent = Event("create",{}) # list of events
        self.lastEvent = self.firstEvent
        self.currentEvent = self.firstEvent  # keeps track of what event an operation is applied to

# Pretty-printing variables:
        self.arguments_print = { 'msg': "", 'from': " apparently from ", 'to': " intended for ", 'info': " ", 'data': "" }
        self.arguments_order = [ 'msg', 'from', 'to', 'info' ]
        self.arguments_type = {'msg': 'str', 'msglist': 'strlist', 'data': 'omit' }
        self.action_preprint = { 'warning': bcolors.WARNING }
        self.action_print = { 'warning': 'warning:' }
        self.action_postprint = { 'warning': bcolors.ENDC+" " }

    # adds event after last event, keeps current event unchanged
    def add_event(self,action,arguments):
#        if self.firstEvent != None:
#            ev = self.firstEvent
#            while ev.next != None:
#                ev = ev.next
#            ev.append(action,arguments)
        if self.lastEvent != None:
            self.lastEvent.append(action,arguments)
            self.lastEvent = self.lastEvent.next
        else:
            # we should never be in this case, since roles are generated with a firstEvent called "create"
            warn("Class: RoleScript, add_event: This should not be happening.")
            self.firstEvent = Event(action,arguments)
            self.currentEvent = self.firstEvent
            self.lastEvent = self.firstEvent

    # set currentEvent to the event it points to
    # returns True on success, False when there is no next event
    def next_event(self):
        if self.currentEvent == None:
            return False
        elif self.currentEvent.next == None:
            return False
        else:
            self.currentEvent = self.currentEvent.next
            return True

    def update_event(self,key,arg):
        self.currentEvent.update_arg(key,arg)

    def update_action(self,action,key,arg):
        ev = self.find_event_action(action,self.firstEvent)
        if ev == None:
            error("No such action in role script")
        evn = self.find_event_action(action,ev.next)
        if evn != None:
            error("Action is not unique")
        ev.update_arg(key,arg)

    def update_first_action(self,action,key,arg):
        ev = self.find_event_action(action,self.firstEvent)
        self.currentEvent = ev
        self.update_event(key,arg)

    def update_next_action(self,action,key,arg):
        ev = self.find_event_action(action,self.currentEvent.next)
        self.currentEvent = ev
        self.update_event(key,arg)

    def exists_next_action(self,action,key,arg):
        if self.find_event_action(action,self.currentEvent.next) != None:
            return True
        else:
            return False

    def set_current_event(self,ev):
        self.currentEvent = ev

    def get_current_event(self):
        return self.currentEvent

    # inserts an event after "currentEvent"
    # if currentEvent is None, an error is raised
    # currentEvent points to newly inserted event
    def insert_event(self,action,arguments):
        if self.currentEvent == None:
            error("No current event to insert into.")
        elif self.currentEvent == self.lastEvent:
            self.add_event(action,arguments)
        else:
            self.currentEvent.insert(action,arguments)
        if not self.next_event():
            error("this should be impossible. Error: I1245")

    # inserts an event after "currentEvent"
    # if currentEvent is None, an error is raised
    # difference to above: currentEvent is not moved
    def insertAfter_event(self,action,arguments):
        if self.currentEvent == None:
            error("No current event to insert into.")
        elif self.currentEvent == self.lastEvent:
            self.add_event(action,arguments)
        else:
            self.currentEvent.insert(action,arguments)

    # as above, but inserts an event before "currentEvent"
    # currentEvent is not moved
    def insertBefore_event(self,action,arguments):
        if self.currentEvent == None:
            error("No current event to insert into.")
        elif self.currentEvent == self.firstEvent:
            warn("inserting before the first event!")
            ev = Event(action,arguments)
            ev.next = self.firstEvent
            self.firstEvent = ev
        else:
            self.currentEvent._insertBefore(action,arguments)
            # the following is necessary because _insertBefore copies data
            # and currentEvent ends up pointing to the previous event
            self.currentEvent = self.currentEvent.next

    def find_event_by_key_arg(self,key,arg,fev=None):
        if fev == None:
            fev = self.firstEvent
        ev = fev
        while ev != None:
            if key in ev.arguments and ev.arguments[key] == arg:
                return ev
            ev = ev.next
        return None

    def find_event_by_key(self,key,fev=None):
        if fev == None:
            fev = self.firstEvent
        ev = fev
        while ev != None:
            if key in ev.arguments:
                return ev
            ev = ev.next
        return None

    def find_event_action(self,key,fev):
        ev = fev
        while ev != None:
            if ev.action == key:
                return ev
            ev = ev.next
        return None

    def add_initial_knowledge(self,t):
        self.IK.add(t)

    def set_name(self,name):
        self.role_name = name

    def output(self,onlyspecial=False,noempty=True):
        print "Role "+self.role_name+":"
        print "  Initial Knowledge: ",
        self.IK.printKnowledge()
        print
        print "  Events:"
        
        ev = self.firstEvent
        while ev != None:
            line = "    "
            if ev.action in self.action_preprint:
                line += self.action_preprint[ev.action]
            if ev.action in self.action_print:
                line += self.action_print[ev.action]
            else:
                line += ev.action
            if ev.action in self.action_postprint:
                line += self.action_postprint[ev.action]
            else:
                line += " "
            for a in self.arguments_order:
                if a in ev.arguments:
                    if a in self.arguments_print:
                        line += self.arguments_print[a]+self.arg2str(ev.arguments,a)
                    else:
                        line += a+": "+str(ev.arguments[a])
            remainder = []
            if not onlyspecial:
                for a in ev.arguments:
                    if a not in self.arguments_order:
                        if (not noempty) or (ev.arguments[a] != ""):
                            if a in self.arguments_print:
                                remainder.append(self.arguments_print[a]+self.arg2str(ev.arguments,a))
                            else:
                                remainder.append(a+": "+self.arg2str(ev.arguments,a))
            if len(remainder) > 0:
                remainder = ",".join(remainder)
                if len(remainder) > 0:
                    remainder = " ("+remainder+")"
            else:
                remainder = ""
            line += remainder
            print line
            ev = ev.next

    def arg2str(self,args,a):
        if a in self.arguments_type:
            t = self.arguments_type[a] 
            if t == "str":
                return args[a]
            elif t == "strlist":
                return ", ".join(args[a])
            elif t == "omit":
                return ""
        return str(args[a])

class Protocol:
    def __init__(self,T,name=""):
        self.protocol_name = name
        self.roles = {}
        self.T = T  # TermAlgebra

    def create_role(self,rolename,terms=[]):
        if rolename != "":
            if rolename not in self.roles:
                self.roles[rolename] = RoleScript(self.T,rolename)

            for t in terms:
                self.roles[rolename].add_initial_knowledge(t)
        else:
            error("Empty rolename.")

    def add_role(self,role,terms=[]):
        if type(role) is RoleScript:
            if role.role_name == "":
                A = "Role"+str(len(self.roles)+1)
                role.set_name(A)
            else:
                A = role.role_name
            if A not in self.roles:
                self.roles[A] = role
            else:
                error("Cannot add role name to protocol. Role name exists.")

            for t in terms:
                self.roles[A].add_initial_knowledge(t)

        elif type(role) is str:
            self.create_role(role,terms)

    def add_event(self,rolename,action,arguments):
                self.roles[rolename].add_event(action,arguments)

    def output(self):
        print "Protocol "+self.protocol_name+":"
        for role in self.roles:
            self.roles[role].output()
            print
        print

####### MAIN ##########
if __name__ == "__nomainhere__":

    R = RoleScript()
    R.add_event("hello","World")
    R.add_event("hello2","World3")
    R.output()

    P = Protocol("P")
    P.add_role(R)
    
    P.output()
#!/usr/bin/env python

#### TODO: Keep track of produced symbols in order to only produce
####   checks for named elements. Keep a list of printed terms (or rather term positions?) 
####  --> We will use symbol_map for this. This requires that 
#### only symbols are named that are actually printed on the screen.
#



class Checks:
    debug = False
    def dbg(self,text):
        if self.debug:
            print text

    def __init__(self,T):
        self.T = T    # T is the TermAlgebra
        self.step = 0 # keeps track of the step in which a term is learned or produced
        self.Dterms = {}  # stores positions that a deconstructed or received term appears in
        self.Cterms = {}  # maps positions to terms (reverse of Dterms)

        self.K = Knowledge(T)
        self.Gamma = [] # as in paper. Collects terms that need to be checked
        self.Delta = {} # Dictionary of derivable term positions. Constructed simultaneously with Gamma, but we never remove elements, we only add them. The key is the message step in which a term was received.
        
        #### The following are for computing a unique representation for
	## known, generated and received terms 
	self.symbol_map = {} # maps positions to names (A position is a list cast to a tuple.) Thus symbol_map[tuple(Dterms[v][0])] should typically give v for a variable v 
	self.symbol_store = [ "alpha", "beta", "gamma", "delta", "epsilon", "zeta", "eta", "theta", "iota", "kappa", "lambda", "mu", "nu", "xi", "omicron", "pi", "rho", "sigma", "tau", "upsilon", "phi", "chi", "psi", "omega" ]
	# remove symbols that have been used as functions
	for fn in self.T.Sigma:
		if fn in self.symbol_store:
			self.symbol_store.remove(f)
	self.used_symbols = {} # this keeps track of used symbols


    # Return a fresh name for expr. (This can be improved
    # by returning a sensible name, e.g. pkA insteadk of alpha, for pk(skA)
    # At the moment expr is not used in the generation of a fresh name
    def fresh_symbol(self,expr):
	    while len(self.symbol_store) > 0:
		    if self.symbol_store[0].lower() in self.used_symbols:
			    self.symbol_store.remove(self.symbol_store[0])
		    else:
			    break
	    if len(self.symbol_store) == 0:
		    # we can do better than the error, but that's for later
		    error("Ran out of symbols!")
	    if self.symbol_store[0].lower() in self.used_symbols:
		    error("chksyntactic.py, fresh_symbol: this should not happen")
	    r = self.symbol_store[0]
	    self.used_symbols[r.lower()] = True
	    self.symbol_store.remove(r)
	    return r

    # give a new name nm to term tm at position l (includes protocol step) 
    # tm is not strictly necessary, can be None
    def rename_expression(self,tm,l,nm):
        self.dbg("Renaming")
        pos = tuple(l)
        self.symbol_map[pos] = nm
        return nm

    # give a name to term tm at position l (includes protocol step) 
    # this is like name_expression, except that the position
    # must not have been assigned a name yet.
    def give_name_expression(self,tm,l):
        self.dbg("(give_name_expression) "+str(tm)+", "+str(l))
        pos = tuple(l)
        if pos in self.symbol_map:
            error("(give_name_expression) name has been assigned before it's "+self.symbol_map[pos])
        elif pos == tuple([-1]):
            error("Asked to give a name to position [-1] for term "+tm.to_str())
        elif tm.is_variable():
            self.dbg("It's a variable")
            s = tm.to_str()
            if s.lower() in self.used_symbols:
                self.dbg("but it's used, creating a new name")
                # the variable that we want to use has been used
                # already. We try to name it by extending with 
                # _0, _1, _2, etc., up to _9
                # If it doesn't work, we'll just give it a 
                # fresh name
                for i in range(10):
                    new_s = s+str(i)
                    if new_s.lower() not in self.used_symbols:
                        s = new_s
                        break
                if s.lower() in self.used_symbols:
                    s = self.fresh_symbol(tm)
            self.used_symbols[s.lower()] = True
            self.symbol_map[pos] = s
            self.dbg("assigning "+s)
            return s
        else:
            s = tm.to_str().replace("(","_").replace(",","_").replace(")","")
            if s.lower() in self.used_symbols:
                s = self.fresh_symbol(tm)
            self.used_symbols[s.lower()] = True
            self.symbol_map[pos] = s
            self.dbg("It's NOT a variable, assigning "+s)
            return s


    # Return the name of the expression at position pos
    # pos is a list
    def get_name_expression(self,pos):
        p = tuple(pos)
        if p in self.symbol_map:
            return self.symbol_map[p]
        else:
            warn("Symbol map: "+str(self.symbol_map))
            error("Position "+str(pos)+" has not been named.")

    # If tm is a variable and named for the first time, 
    # return its name and mark it as used. 
    # Else give the expression a fresh name.
    def name_expression(self,tm,l):
            if tm != None:
                self.dbg("Naming "+tm.to_str()+" at position "+str(l))
                pass
            else:
                self.dbg("Answering query for position "+str(l))
                self.get_name_expression(l)
	    pos = tuple(l)
	    if pos in self.symbol_map:
		    self.dbg("Seen it before, it's "+self.symbol_map[pos])
		    return self.symbol_map[pos]
            elif pos == tuple([-1]):
                warn("Asked to give a name to position [-1] for term "+tm.to_str())
                return tm.to_str()

            else:
                return self.give_name_expression(tm,l)

    # return True if annotation ann has been assigned a name
    def has_name(self,ann):
        if tuple(ann) in self.symbol_map:
            return True
        return False

    # return the largest prefix of ann that occurs in symbol_map
    # this is the lowest, named term that is a super-term of the 
    # term at ann
    # returns [] if there is no such position
    def lowest_named_position(self,ann):
        if tuple(ann) in self.symbol_map:
            return ann
        lowest = []
        for l in self.symbol_map:
            ell = list(l)
            if prefix(ell,ann) and prefix(lowest,ell):
                lowest = ell
        return lowest

    # Name a message term m in position ann
    # based on presently derivable subterms given in the distionary self.Delta
    # and constructible knowledge.
    # ann is prefixed with a step number.
    def name_by_derivable(self,m,ann):
        pos = ann[1:]
        self.dbg("(name_by derivable) naming "+m.to_str()+" at "+str(ann))
        self.dbg("(name_by_derivable) Delta for this step: "+str(self.Delta[ann[0]]))
        # If there are positions that pos is a strict prefix of 
        # (i.e. we can derive a subterm of this position)
        if any([ prefix(pos,p) for p in self.Delta[ann[0]] if p != pos]):
            # then we apply this naming procedure to all 
            # children of this position and combine the expressions
            l = m.argument_positions(pos)
            if len(l) > 0:
                s = [ self.name_by_derivable(m, [ann[0]]+i) for i in l ]
                self.dbg("s = "+str(s))
            # the following assures us that we can name this position
            # self.give_name_expression(m.position(pos[1:]),pos)
            # now we give the actual name we want
                self.dbg("renaming "+str(ann))
                return self.rename_expression(None,ann,m.position(pos).top_name()+"("+",".join(s)+")")
            else:
                self.dbg("Naming "+str(ann))
                return self.name_expression(m.position(pos),ann)
        else:
            # else we check whether this position is constructible
            is_c = self.is_constructible_with_name(m.position(pos),ann)
            if is_c == True:
                self.dbg("Constructing "+str(ann))
                return self.construct(m.position(pos),ann)
            elif is_c != False:
                warn("(name_by_derivable) Need to give a fresh name!\n      This part of the code has not been stress-tested.")
                self.dbg("(name_by_derivable) Constructing "+m.position(pos).to_str()+" at "+str(ann))
                return self.construct(m.position(pos),ann)                
            else:
                # give a name to this position, since 
                # neither it nor its subterms are derivable/constructible
                self.dbg("Naming "+str(ann))
                return self.name_expression(m.position(pos),ann)


    # Rename a message m by giving new names to those positions
    # that have underivable subterms. Returns true if at least one subterm is derivable.
    def rename_by_derivable(self,m,ann):
        self.dbg("(rename_by_derivable) Arg: m = "+m.to_str()+", ann = "+str(ann))
        r = True
#        if ann in self.Gamma: 
#            self.dbg("ann = "+str(ann)+" is in self.Gamma")
#            s = False
#        else:
#            s = self.construct(m.position(ann[1:]),ann) 
        if self.is_constructible(m.position(ann[1:]),ann) == False:
            l = m.argument_positions(ann[1:])
            r = any([self.rename_by_derivable(m,[ann[0]]+pos) for pos in l])
            if not r:
                newterm = self.name_expression(m.position(ann[1:]),ann)
                m.replace_in_pos(ann[1:],Term(self.T,newterm))
                self.dbg("(rename_by_derivable) Replacement at position "+str(ann[1:])+": newterm = "+newterm)
        else:
            s = self.construct(m.position(ann[1:]),ann) 
            self.dbg("(rename_by_derivable) s = "+str(s))
            self.dbg("(rename_by_derivable) Replacement at top level: "+str(s))
            m.replace_in_pos(ann[1:],Term(self.T,s))
        self.dbg("(rename_by_derivable) returning "+str(r))
        return r

    # map terms to positions and vice versa
    def annotate(self,m,l):
        s = m.to_str()
        if s in self.Dterms:
            if l not in self.Dterms[s]:
                self.Dterms[s].append(l) 
        else: 
            self.Dterms[s] = [l]
        if tuple(l) in self.Cterms:
            warn("Repeated annotation of term at position "+str(l))
        self.Cterms[tuple(l)] = m

    # ann is a newly derivable term. Find parent that is named, produce an equation that
    # explains how ann can be computed.
    def produce_computation(self,ann):
        self.dbg("(produce_computation) ann ="+str(ann))
        pos = self.lowest_named_position(ann)
        self.dbg("pos is "+str(pos)+" which is equal to "+self.get_name_expression(pos)+" literally "+self.Cterms[tuple(pos)].to_str())
        if pos == ann:
            self.dbg("This term is already known.")
            return ""
        else:
            prevname = self.get_name_expression(pos)
            newname = self.name_by_derivable(self.Cterms[tuple([pos[0]])],pos)
        return prevname+" = "+newname


    # checks whether syntactic/non-syntactic checks can be produced at the present protocol step
    # for message m annotated with ann  (ann is the ghost position)
    def produce_check(self,m,ann,step):
        is_c = self.is_constructible_with_name(m,ann)
        if is_c == True:
            check = self.construct(m,ann)
            if ann[0] == step:  # if the term we're checking has been learned in this round
                self.dbg("(produce_check) syntactic check: "+m.to_str()+"@"+str(ann)+" = "+check)
                self.rename_expression(None,ann,check)
                return []
            else:
                self.dbg("(produce_check) verification check: "+m.to_str()+"@"+str(ann)+" = "+check)
                ### Must check that ann has a name. If not, 
                ### must find lowest ancestor that does and make
                ### a comparison to that ancestor first.
                self.dbg("have a check for :"+self.name_by_derivable(self.Cterms[tuple([ann[0]])],ann))
                if self.has_name(ann):
                    prevname = self.name_expression(None,ann)
                    # Should we rename in this case?
                    self.rename_expression(None,ann,check)                    
                    return [ prevname+" = "+check ]
                else:
                    error("This should not happen anymore")

        elif is_c != False:
            # must find out where the unnamed term stems from, produce a check for that position
            ### give name and produce a "check" that indicates the naming, call produce_check again
            warn("(produce_check) "+m.to_str()+" at "+str(ann)+" is constructible, but without name")
            error("This should not happen anymore")
            print "Inconstructible positions: ",is_c
            for p in is_c:
                pos = self.lowest_named_position(p)
                print "Constructing for "+str(pos)+" which is "+self.get_name_expression(pos)+" literally "+self.Cterms[tuple(pos)].to_str()
                prevname = self.get_name_expression(pos)
                newname = self.name_by_derivable(self.Cterms[tuple([pos[0]])],pos)
                print "(produce_check) Finished naming: "+prevname+" is "+newname
                check = self.produce_check(m,ann,step)
                if check != False:
                    return [ prevname+" = "+newname ] + check
                else:
                    return [ prevname+" = "+newname ] 
            error("Stopping here, must implement this first")
            s = self.construct(m,ann)
            return s
        else:
            return False


    def __call__(self,ABC):
        self.check_protocol(ABC.prot)

    def check_protocol(self,prot):
        for r in prot.roles:
	    self.__init__(prot.T)
	    self.check_role(prot.roles[r])

## Go through the protocol specification and add checks
    def check_role(self,role):
       #firstWarning = True # this will output the deriviation rules only once
#       role.arguments_print['gamma']="^"
#       i = role.arguments_order.index('msg')
#       role.arguments_order = role.arguments_order[:i+1]+['gamma']+role.arguments_order[i+1:]
       role.arguments_order.append('msglist')

       role.arguments_print['eq'] = ""        
       role.arguments_print['current knowledge'] = "" 
       role.arguments_print['msglist']=""

       role.arguments_type['eq'] = "str"
       role.arguments_type['current knowledge'] = "strlist"

       role.arguments_order += ['eq','current knowledge']
       role.action_print['knows'] = 'role knows:' 

       self.dbg("Role "+role.role_name)
       role.update_action("create",'msglist',[])

       # Since the user-specified initial knowledge may not be minimal
       # we first collect the knowledge, then compute the naming for 
       # the role script
#       for m in role.IK.list():
#	       self.knows(m,nostep=True)

       role.IK.closeKD()
       arg = role.firstEvent.arguments['msglist']
       for m in role.IK.minimalKnowledgeList():
           self.dbg("Running knows now")
#           tm = self.knows(m,itself=True)  #### perhaps better, once the
                                            #### itself keyword is fixed
           tm = self.knows(m)
#           print "I know: "+self.name_expression(tm,[self.step])
#           n = self.construct(Term(self.T,m),[-1])  ### was self.step+1, we know this term, as it is in the minimal knowledge. -1 indicates that it is independent of order of received terms
#           self.dbg("n = "+str(n))
           arg.append(m)
       role.update_action("create",'msglist',arg)

       role.set_current_event(role.firstEvent)

       step = 0
       while role.next_event(): # this skips the create event
	   step += 1
           currEvent = role.currentEvent
	   action = role.currentEvent.action
	   arg = role.currentEvent.arguments
#	   if ev.next == None:
#		   ev.insert("unchecked",self.Gamma)
#		   ev = ev.next
	   if action == "receive":
                m = arg['msg']
                self.dbg("Receiving: "+m)
                mrenamed,list_of_checks,list_of_computations = self.learn(m)
		role.currentEvent.add_arg("gamma",self.step)
#		role.insert_event("knows",{'current knowledge': self.K.minimalKnowledgeList()})
		for check_term in list_of_computations:
			role.insert_event("compute",{'eq': check_term})
			self.dbg("Added computation: "+check_term)
		for check_term in list_of_checks:
			role.insert_event("check",{'eq': check_term})
			self.dbg("Added check: "+check_term)
                if self.debug: 
                    self.printGamma()
                    self.printDelta()
                    role.insert_event("dbg topGamma",{'msglist': self.listGamma(self.topGamma())})
                    role.insert_event("dbg minimalGamma",{'msglist': self.listGamma(self.minimalGamma())})
                    role.insert_event("dbg annotated minimal Knowledge",{'msglist': self.listMinimalAnnotatedKnowledge()})
                m = Term(self.T,m)
                if self.debug:
                    role.insert_event("dbg learned term as string",{'str': m.to_str()})
                    role.insert_event("syntactic receive",{'str': m.to_str()})
                # self.rename_by_derivable(m,[self.step])
                self.dbg("2 more steps in 'receive'")
		role.insert_event("knows",{'current knowledge': self.listMinimalKnowledge()})
                # currEvent.update_arg('msg', m.to_str())
                currEvent.update_arg('msg', mrenamed)

	   elif action == "fresh":
                m = arg['msg']
                self.dbg("Fresh: "+m)
		self.knows(m)
		role.currentEvent.add_arg("gamma",self.step)
                currEvent.update_arg('msg', self.get_name_expression([self.step]))
	   elif action == "send":  
                m = Term(self.T,arg['msg'])
                self.dbg("Sending: "+m.to_str())
                self.dbg("Initial knowledge:")
                self.dbg(" ".join([i for i in self.K.KD]))
                self.dbg("")
                if self.K.is_constructible(m):
                    self.dbg(m.to_str()+" is derivable")
                else:
#			if firstWarning:
#				print "Derivation rules: "
#				self.K.T.printDrules()
#				firstWarning = False
#			print "Role knowledge: "
#			self.K.printKnowledge()
                    warn("Role "+role.role_name+", step "+str(step)+", send event: term "+m.to_str()+" is not derivable")
                    role.insertBefore_event("warning",{'msg':m.to_str(),'info':"not derivable",'knowledge':self.K.list()})
#			warn("Role "+role.role_name+": "+m.to_str()+" is not derivable")
                self.rename_by_derivable(m,[-1]) ### there is no natural step number associated to this term, since we are sending
                if self.debug:
                    role.insert_event("syntactic send",{'str': m.to_str()})
                currEvent.update_arg('msg', m.to_str())
	   else:
                if 'msg' in arg:
			m = arg['msg']
		else: 
			m = str(arg)
		role.insertBefore_event("warning",{'info':"unknown action '"+action+"' with argument "+m})
                self.dbg("Closed knowledge:")
                self.dbg(" ".join([i for i in self.K.KD]))
                self.dbg("")

       if len(self.Gamma) > 0:
	       role.action_print['not checked'] = 'role cannot check:' 
#	       role.add_event("Gamma",{'Gamma': self.listGamma()})
	       role.add_event("not checked",{'msglist': self.listMinimalGamma()})
#       print "Remaining elements in Gamma for role "+role.role_name+":"
#       self.printGamma()


    def closeKDG(self):
        again = True
        while again:
            again = False
            for m in self.K.KD.keys():
                t = self.K.KD[m]
#                self.dbg("Analyzing: "+t.to_str()+" which is "+m)
		if not t.is_variable():  # we cannot deconstruct vars
                    # should we assert that self.K.T == self.T ?
                  for r in self.T.drules:
                    self.dbg("-- Applying rule "+self.T.rule2str(r)+" to "+m)
                    u = mgusub(self.T,m,r['ld'].to_str())
                    self.dbg("-- Result is "+str(u))
                    if unifies(u):
                        l = []
                        for rt in r['lc']:
                            l.append(rt.apply(u))
                        rhs = r['r'].apply(u)
                        rhspositions = []
                        for i in self.Dterms[m]:
                            rhspositions.append(i+r['rdp'])
                        # need to find all positions rhs.to_str() could have 
                        # and test whether one of them is missing
                        if rhs.to_str() not in self.K.KD:
                            if self.K.klistup(l):  
                                self.K.add(rhs)
                                for i in rhspositions:
                                    self.annotate(rhs,i)
                                    self.dbg("1 Adding "+rhs.to_str()+"@"+str(i))
                                    self.Gamma.append([rhs,i])
                                    self.Delta[i[0]].append(i[1:])
                                again = True
                        else:
                            for i in rhspositions:
                                if i not in self.Dterms[rhs.to_str()]:
                                    self.annotate(rhs,i)
                                    self.dbg("2 Adding "+rhs.to_str()+"@"+str(i))
                                    self.Gamma.append([rhs,i])
                                    self.Delta[i[0]].append(i[1:])

    # similar to K.is_constructible, but takes message positions into account
    def is_constructible(self,t,ann):
        s = t.to_str()
        if t.is_variable():
            if s in self.K.KD:
                for i in self.Dterms[s]:
                    if not prefix(ann,i):
                            return True
            return False
        else:
            # try to construct it from pieces:
            if s in self.K.KD:
                for i in self.Dterms[s]:
                    if not prefix(ann,i):
                        return True
            for i in t.arguments():
                # should ann be modified here?
                # I don't think so, else e.g. the pair(a,a) at [x]
                #  can be constructed from itself 
                # with the two a's swapped: pair(a^[x,2],a^[x,1]) 
                res = self.is_constructible(i,ann)
                if res == False:
                    return False
            return True


    # same as is_constructible, but checks whether the term has
    # a name assigned
    # returns True, False, or a list, meaning, respectively
    # that a term is constructible and has a name
    # is not constructible,
    # is constructible, but without a name at the positions in the list
    def is_constructible_with_name(self,t,ann):
        self.dbg("(is_constructible_with_name) t = "+t.to_str()+" ann = "+str(ann))
        s = t.to_str()
        if s in self.K.KD:
            l = [ i for i in self.Dterms[s] if not prefix(ann,i) ]
            lnamed = [ i for i in l if self.has_name(i) ]
            for i in lnamed:
                return True
            for i in l:
                return [ i ]
            return False
        if t.is_variable():
            return False
        else:
            l = [ self.is_constructible_with_name(i,ann) for i in t.arguments()]
            self.dbg("(is_constructible_with_name) l = "+str(l))
            if False in l:
                return False
            elif all([ i == True for i in l]):
                return True
            else:
                return [ i for i in l if i != True ]

    # Return constructed string, assign names where necessary
    # THIS IS ONLY TO BE USED WHEN THE STRING IS GOING TO OCCUR IN THE PROTOCOL
    # because later terms might make a reference to the substrings produced 
    def construct(self,t,ann,noname=False,pos=None):
        if pos == None:
            pos = ann
        self.dbg("(construct) called for annotation "+str(ann)+" at position "+str(pos))
        s = t.to_str()
        if s in self.K.KD:
            self.dbg("(construct) the term at position "+str(ann)+" is in K.KD")

            l = [ i for i in self.Dterms[s] if not prefix(ann,i) ]
            lnamed = [ i for i in l if self.has_name(i) ]
            for i in lnamed:
                self.dbg("(construct) returning the name at position "+str(i)+" for annotation "+str(ann))
                return self.get_name_expression(i)
            if noname:
                warn("Symbol map: "+str(self.symbol_map))
                error("(construct) Couldn't get name for expression "+s)
            else:
                for i in l:
                    # i.e. if tuple(i) not in self.symbol_map for any i
                    warn("(construct) Ann = "+str(ann)+": This is an unnamed symbol: "+s+"^"+str(i))
                ##### The big question here is what to return:
                    # 1. create a fresh name for position i? +++yes+++
                    # 2. create a fresh name for position pos?
                    # 3. return the string s and 
                    #   a) assign the string s to position i?
                    #   b) assign the string s to position pos?
                    #return s+"^{"+str(i)+"}"
                    ##### We could give both a name
                    return self.name_expression(t,i)
#                    return s # +"^"+str(i)+""
#                    return self.name_expression(t,pos)
        if t.is_variable():
            error("Attempt to construct an underivable term at position "+str(ann))
        else:
            # try to construct it from pieces:
            r = t.top_name()+"("
            for i in t.argument_positions():
                res = self.construct(t.position(i),ann,noname,pos+i)
		if r[-1] != "(": # check whether this is the first argument to be added
			r += ","+res # it is not, thus put a comma to separate from prev arg
		else:
			r += res # it's the first arg, no preceding comma necessary
            r += ")"
            return r
    
    # we return the largest elements of Gamma which have not been checked
    # because some of their subterms have not been checked (and the subterms
    # appear in Gamma)
    def topGamma(self):
	    top = []
	    for gi in self.Gamma:
		    noprefix = True
		    for gj in self.Gamma:
			    if gi != gj and prefix(gj[1],gi[1]):
				    noprefix = False
				    break
		    if noprefix: top.append(gi)
	    return top

    # we return the smallest elements of Gamma which have not been checked
    def minimalGamma(self):
	    minimal = []
	    for gi in self.Gamma:
		    noprefix = True
		    for gj in self.Gamma:
			    if gi != gj and prefix(gi[1],gj[1]): # note the order
				    noprefix = False
				    break
		    if noprefix: minimal.append(gi)
	    return minimal

    def printGamma(self):
        print "Gamma Terms:"
        for i in self.Gamma:
            print "  "+i[0].to_str()+"@"+str(i[1])

    def printDelta(self):
        print "Delta Terms:"
        for i in self.Delta:
            print str(i)+": "+str(self.Delta[i])
    
    def listGamma(self,Gamma=None):
        if Gamma == None:
            Gamma = self.Gamma
#        return [ i[0].to_str()+"@"+str(i[1]) for i in Gamma ]
        return [ self.get_name_expression(i[1]) for i in Gamma ]

    def listMinimalGamma(self):
#	    return [ i[0].to_str()+"@"+str(i[1]) for i in self.minimalGamma() ]
	    return [ self.get_name_expression(i[1]) for i in self.minimalGamma() ]

    def listMinimalAnnotatedKnowledge(self):
        return [ i+"@"+str(self.Dterms[i]) for i in self.K.minimalKnowledgeList() ]

    def listMinimalKnowledge(self):
#        warn("Incorrect implementation of minimal Knowledge.")
# This should be:
        return [ self.construct(i,[-1],noname=True) for i in self.K.minimalKnowledge() ]
# but it doesn't work for example6 at the moment. Needs a fix.
#        return [ self.construct(i,[-1]) for i in self.K.minimalKnowledge() ]


    def learn(self,s):
	    list_of_checks = []
            priorKnowledge = []
            for i in self.Delta:
                if i != self.step:
                    priorKnowledge += [ [i]+l for l in self.Delta[i] ]

            self.dbg("Prior Knowledge: "+str(priorKnowledge))
            self.knows(s,learned=True) # True indicates that it has been learned, causes Gamma.append(...)

	    Gammaprime = []
            # Do syntactic checks first
	    for ghost in self.Gamma:
                if ghost[1][0] == self.step: 
                    check = self.produce_check(ghost[0],ghost[1],self.step) # ghost[0] is the message term, ghost[1] the position (annotation)
                else:
                    check = False
                if check == False:
                    Gammaprime.append(ghost)
                else:
                    list_of_checks += check
	    self.Gamma = Gammaprime
            Gammaprime = []

            # we name the received term based on its derivable positions 
            self.dbg("(learn) Derivable positions: "+str(self.Delta[self.step]))
            r = self.name_by_derivable(Term(self.T,s),[self.step])

            # Before doing the next checks, write down the previously received terms 
            # that we can now interpret in more detail
            #### Perhaps we should do this in a specific order
            list_of_computations = []
            currentKnowledge = []
            for i in self.Delta:
                if i != self.step:
                    currentKnowledge += [ [i]+l for l in self.Delta[i] ]
            self.dbg("Current Knowledge: "+str(currentKnowledge))
            for i in [ j for j in currentKnowledge if j not in priorKnowledge ]:
                comp = self.produce_computation(i)
                if comp != "":
                    list_of_computations.append(comp)

            # Now do all checks 
	    for ghost in self.Gamma:
                check = self.produce_check(ghost[0],ghost[1],self.step) # ghost[0] is the message term, ghost[1] the position (annotation)
                if check == False:
                    Gammaprime.append(ghost)
                else:
                    list_of_checks += check
	    self.Gamma = Gammaprime

# must store this representation somewhere, so that we can come back to it
# later
            ### Name the minimal elements in Gamma:
#            self.dbg("(learn) Arg: "+s+" naming minimal elements in Gamma...")
#            for ghost in self.minimalGamma():
#                self.name_expression(ghost[0],ghost[1])
            ## constructing this learned term should now be possible from those elements
	    return r,list_of_checks,list_of_computations

    def knows(self,s,**kwargs):
        m = Term(self.T,s)
        self.K.add(m)
        if 'nostep' not in kwargs:
	    self.step += 1
            self.dbg(str(self.step)+". Knows/learned: "+m.to_str())
	    self.annotate(m,[self.step])
        else:
            self.dbg('(knows) with nostep keyword for '+m.to_str())

        # naming and gamma policy depends on third parameter
        if 'learned' in kwargs: 
            self.dbg("(knows) b/c learned")
            self.Gamma.append([m,[self.step]])
            self.Delta[self.step] = [[]]
        elif 'noname' in kwargs or 'nostep' in kwargs:
            self.dbg("(knows) no naming")
            pass 
        elif 'name' in kwargs:
            self.dbg("(knows) name = "+kwargs['name'])
            self.rename_expression(m,[self.step],kwargs['name'])
        elif 'itself' in kwargs:
            self.dbg("(knows) name = "+s+" (itself)")
            #### FIXME: TODO: this should check whether Dterms[s] exists
            #### Then find a name used by a position in Dterms[s]
            #### and return that name.
            self.rename_expression(m,[self.step],s)
        else: 
            self.dbg("(knows) (no keywords given)")
            self.name_expression(m,[self.step])
        self.closeKDG()  # this will perform further Gamma.appends

        return m

####### MAIN ##########
if __name__ == "__main__":
    
# Choose input format
   arguments = sys.argv[1:]

# Choose tamarin output format?
#   ABC = tamarin.Tamarin()
   ABC = AnB()
#   ABC = strand()

## Parse the specification into a protocol type
   ABC.read_prot(arguments)

## Output the protocol
#   ABC.output()


   check = Checks(ABC.T)
   check(ABC)

   ABC.output()

