#!/usr/bin/env python

from syntactic import *

#explanations = True
explanations = False

# return create event data stored in role script
def create_data(arg):
    if 'msglist' in arg:
        return arg['msglist'][:]
    elif 'data' in arg:
        return [ p[0] for p in arg['data'] ]
    else:
        error("Create event contains no known data types")

### transition labels (actions)
def ruletransition(rolename,num,actionlist,terms,lemmas):
    rolestep = ""
    if 'functionality' in lemmas:
        rolestep = rolename+str(num)+"("+",".join(terms)+")"
        if rolename in lemmas['functionality']:
            lemmas['functionality'][rolename].append(rolestep)
        else:
            lemmas['functionality'][rolename] = [rolestep]
    return "]\n --[ " + ", ".join([rolestep]+actionlist) + " ]->\n ["

def printlemmas(lemmas):
    # Todo: make sure that the time point variables do not conflict with other
    # variables used in a lemma
    if 'functionality' in lemmas:
        print """lemma functionality: exists-trace 
  "
   (All #i #j. Setup()@i & Setup()@j ==> #i = #j)"""
        exists = []
        facts = []
        for rn in lemmas['functionality']:
            n = len(lemmas['functionality'][rn])
            prevtimepoint = ""
            for i in range(n):
                timepoint = "#c"+rn+str(i)
                exists.append(timepoint)
                facts.append(lemmas['functionality'][rn][i]+"@"+timepoint)
                if i > 0:
                    facts.append(prevtimepoint+" < "+timepoint)
                prevtimepoint = timepoint
        if len(exists) > 0:
            print "   & Ex "+" ".join(exists)+". "+" & ".join(facts)+'\n"'
        else:
            print '"'
# Example: 
#  & Ex #i0 #i1 #i2 #j0 #j1. A0()@i0 & A1()@i1 & A2()@i2 & B0()@j0 & B1()@j1 & i0 < i1 & i1 < i2 & j0 < j1"

def printaxioms(axioms):
    if 'equality' in axioms:
        print """axiom equality: 
  "All a b #i. Eq(a,b)@i ==> a = b"
"""


def header(name):
    theoryname = keep_char(name,alphanum+"_")
    if theoryname == "":
        theoryname = "noname"
    r = """theory """+theoryname+"""
begin
"""
    if explanations:
        r += """
/* 

// E.g.:
functions: h/1, aenc/2, adec/2, pk/1
equations: adec(aenc(m, pk(k)), k) = m
           // (equations must be subterm convergent)

builtins: // diffie-hellman, signing, asymmetric-encryption, 
//   symmetric-encryption, hashing 

Variables:
    ~x   denotes  x:fresh
    $x   denotes  x:pub
    #i   denotes  i:temp
    i    denotes  i:msg
    'c'  denotes a public name 'c \in PN'; i.e., a fixed, global constant
*/
"""
    return r

def footer():
    if not explanations:
        r = ""
    else:
        r = """
/*
  All      for universal quantification, temporal variables are prefixed with #
  Ex       for existential quantification, temporal variables are prefixed with #
  ==>      for implication
  &        for conjunction
  |        for disjunction
  not      for  negation

  f @ i    for action constraints, the sort prefix for the temporal variable 'i'
           is optional

  i < j    for temporal ordering, the sort prefix for the temporal variables 'i'
           adn 'j' is optional

  #i = #j  for an equality between temporal variables 'i' and 'j'
  x = y    for an equality between message variables 'x' and 'y'

Example Axiom (used to restrict set of considered traces):

axiom Equality_Checks_Succeed: "All x y #i. Eq(x,y) @ i ==> x = y"

Example lemma:

lemma Client_session_key_secrecy:
  " /* It cannot be that a  */
    not(
      Ex S k #i #j.
        /* client setup a session key 'k' with a server'S' */
        SessKeyC(S, k) @ #i
        /* and the adversary knows 'k' */
      & K(k) @ #j
        /* without having performed a long-term key reveal on 'S'. */
      & not(Ex #r. LtkReveal(S) @ r)
    )
  "
*/

"""
    return r+"end\n"

## This is the output function:
class Tamarin(AnB):
    debug = False
    def dbg(self,text):
        if self.debug:
            print text

    def agstate(self,rolename,num,terms,with_public=False):
        if num == 0:
            return ", ".join(terms) # OLD: "!Ag_St_"+rolename+str(num)+"("+", ".join(terms)+")"
        else:
            if with_public:
                roles = self.prot.roles.keys()
                public_terms = dict(zip(roles,[ "$"+r for r in roles]))
                terms = [ assign(self.T,t,public_terms) for t in terms ]
                self.dbg("(agstate) roles = "+str(roles)+" public terms are "+str(public_terms)+" terms = "+str(terms))
            return "Ag_St_"+rolename+str(num)+"("+", ".join(terms)+")"

## Rewriting system
   # Left-hand side of send rule
    def lhsrule(self,A,num,seen,messagelist,letlist):
    # if num == 0:
    #   return "rule "+A+"_"+str(num+1)+":\n [ "+", ".join(messagelist)
    # else:
        msglist = messagelist+[ self.agstate(A,num,seen) ]
        letstatement = "\n    ".join(letlist)
        if letstatement != "":
            letstatement = "  let\n"+letstatement+"\n  in\n"
        return "rule "+A+"_"+str(num+1)+":\n"+letstatement+" [ "+", ".join(msglist)

   # Right-hand side of send rule
    def rhsrule(self,A,num,seen,messagelist):
        # the last argument in agstate should be true when public terms
        # are to be decorated with $. This is the case when the LTK's are
        # consumed.
        msglist = messagelist+[ self.agstate(A,num+1,seen,num == 0) ]
        return ", ".join(msglist)+" ] \n\n"


    def print_functions(self):
        if len(self.T.Sigma) > 0:
            print "functions: ",
        l = []
        for f in self.T.Sigma:
            l.append(f+"/"+str(self.T.Sigma[f]))
        print ", ".join(l)

    def print_equations(self):
        l = []
        if len(self.T.Sigma) > 0: 
            print "equations: ",
        for f in self.T.Equations:
            l.append(f.replace("%",""))
        print ",\n    ".join(l)


    # collect all the initial knowledge to determine the setup rules
    # return a list of fresh terms and a dictionary mapping 
    # rolename to initial knowledge of role
    def compute_fresh_known(self):
        roles = self.prot.roles

        publicterms = []
        freshterms = []
        knownterms = {}

        for role in roles:
            # collect all terms a role initially knows in knownterms[role]
            knownterms[role] = []

            event = roles[role].firstEvent
            rolename = roles[role].role_name
            # if there is a create event we use that for the initial knowledge
            if event.action in ["create"]:
                self.dbg("Using create event")
                IKlst = create_data(event.arguments)
            else: # else we use IK.KD
                self.dbg("NOT using create event")
                IKlst = [ t for t in roles[role].IK.KD ]

### WE have a BUG: public SUB-terms in IKlst are not marked
### Fix the bug by substitution in the message term t.
            for t in IKlst:
                if t in roles:
                    # role names are public terms
                    publicterms.append("$"+t)
                    knownterms[role].append("$"+t)                    
                elif t in roles[role].IK.KD:
                    if roles[role].IK.KD[t].is_variable():
                        # other variables are fresh terms
                        freshterms.append(t)
                        knownterms[role].append(t)
                    else:
                        # all atomic subterms are fresh unless they're a role name
                        for tv in roles[role].IK.KD[t].vars():
                            if tv not in roles:
                                freshterms.append(tv)
# NEW:
                        knownterms[role].append(t)
# OLD:                               knownterms[role].append(tv)

        return freshterms,knownterms

    def ltk_facts(self,freshterms,knownterms):
        roles = self.prot.roles
        ltk = {} # maps term to roles that can derive the term from their 
                 # initial knowledge. 
        ltkfacts = {} # maps role name to ltk facts that the role receives
        all_ltk_facts = [] # collects all facts
        c = 0 # counts different ltk facts
        for role in roles:
            for t in knownterms[role]:
                if "$" not in t:
                    if t not in ltk:
                        c += 1
                        # Add the names of all roles that can derive this term
                        # to the LTK fact
                        ltkterms = sorted([ "$"+r for r in roles if roles[r].IK.is_derivable(Term(self.T,t)) ])
                        # Name the Ltk fact with a unique index (counter c)
                        ltkterms = "!Ltk_"+str(c)+"("+", ".join(ltkterms)+", "+t+")"
                        ltk[t] = ltkterms
                        all_ltk_facts.append(ltkterms)

            # the best thing to do instead of creating all these
            # different Ltk facts above would be to find symmetries in
            # ltk and create simpler rules that produce them. This would
            # reduce the number of differently indexed ltk_i facts and 
            # make a simpler theory file. TODO.

        for role in roles:
            ltkfacts[role] = [ ltk[t] for t in knownterms[role] if t in ltk ]

        ltkfacts["__all_ltk_facts__"] = all_ltk_facts

        return ltkfacts

    def print_setup_rules(self):
        roles = self.prot.roles
        freshterms, knownterms = self.compute_fresh_known()
        freshterms = list(set(freshterms))
        ltkfacts = self.ltk_facts(freshterms,knownterms)
        self.dbg("LTKFACTS: "+str(ltkfacts))

        freshfacts = [ "Fr("+t+")" for t in freshterms ] 

        print "rule Setup:"
        print " [ "+", ".join(freshfacts),
        print " ]\n--[ Setup() ]->\n [ ",
### OLD:
#        l = [ agstate(role,0,knownterms[role]) for role in roles ]
#        print ", ".join(l),
#        print "]"
#        return knownterms
### NEW:
        print ", ".join(ltkfacts["__all_ltk_facts__"]),
        print "]"
        return ltkfacts,knownterms

    def output(self):

        print header(self.prot.protocol_name)
        self.print_functions()
        self.print_equations()
        print

        # simplest solution to assign to knownterms, because it will 
        # change later
        ltkfacts,knownterms = self.print_setup_rules()

        print

        roles = self.prot.roles

        letstatement = []
        lemmas = { 'functionality': {}}
        axioms = {}

# now create rules for each role script
        for role in roles:
            event = roles[role].firstEvent
            rolename = roles[role].role_name

            counter = 0
            premises = {} 
            consequences = {} 
            actions = []
#            seen = []
#            seen_now = []
# OLD:            previousAgentState = [ t.replace("$","") for t in knownterms[role] ]
# NEW:
            previousAgentState = ltkfacts[role]
            newAgentState = []
#            knowledge_map = {}
            outputnow = False

            while event != None:
            # collect knowledge
                if event.action in ["create"]:
                    newAgentState += create_data(event.arguments)
#                        knowledge_map[p[0]] = p
#                        knowledge_map[p[0]+"^"+str(p[1])] = p
                    outputnow = True
                    consequences[event.action] = create_data(event.arguments)
                if event.action in ["fresh"]:
                    p = (event.arguments['msg'],event.arguments['gamma'])
#                    knowledge_map[p[0]] = p
#                    knowledge_map[p[0]+"^"+str(p[1])] = p
                    newAgentState.append(p[0])
            # collect premises
                if event.action in ["fresh","receive"]:
                    if event.action in premises:
                        premises[event.action].append(event.arguments['msg'])
                    else:
                        premises[event.action] = [event.arguments['msg']]
                if event.action in ["knows"]:
                    newAgentState += event.arguments['current knowledge']
                if event.action in ["send"]:
                    outputnow = True
                    if event.action in consequences:
                        consequences[event.action].append(event.arguments['msg'])
                    else:
                        consequences[event.action] = [event.arguments['msg']]
#                if event.action in ["learn","fresh"]:
#                    seen_now += [ event.arguments['msg'] ]
            # collect computation steps and add to "let" construct:
                if event.action in ["compute"]:
                    letstatement.append(event.arguments['eq'])
            # collect checks and add to rule actions:
                if event.action in ["check"]:
                    axioms['equality'] = True
                    actions.append("Eq("+event.arguments['eq'].replace("=",",")+")")
            # set output signal if this is the last event
                if event.next == None:
                    outputnow = True
            # output if signal is set
                if outputnow:
                    newAgentState = list(set(newAgentState))
                    newAgentState.sort()

                    outputnow = False
                    premmsglist = []
                    consmsglist = []

                    if "fresh" in premises:
                        premmsglist+=["Fr("+x+")" for x in premises['fresh']]
                    if "receive" in premises:
                        premmsglist+=["In("+x+")" for x in premises['receive']]
                    if "send" in consequences:
                        consmsglist+=["Out("+x+")" for x in consequences['send']]

                    print self.lhsrule(role,counter,previousAgentState,premmsglist,letstatement), 
                    print ruletransition(role,counter,actions,[],lemmas),
                    print self.rhsrule(role,counter,newAgentState,consmsglist)

                    previousAgentState = newAgentState[:]
                    premises = {}
                    consequences = {}
                    counter += 1

                event = event.next

        printaxioms(axioms)
        printlemmas(lemmas)

        print footer()

####### MAIN ##########
if __name__ == "__main__":
    
# Choose input format
#   from strand import *
#   from chksyntactic import *
   arguments = sys.argv[1:]

   ABC = Tamarin()

## Parse the specification into a protocol type
   ABC.read_prot(arguments)

## Output the protocol
#   ABC.output()

   check = Checks(ABC.T)
   check(ABC)

   ABC.output()

